@@ -24,6 +24,7 @@
 #include <stdarg.h>
 #include <getopt.h>
 #include <fcntl.h>
+#include <linux/limits.h>
 
 #include "common/utils.h"
 #include "kerncompat.h"
@@ -32,10 +33,12 @@
 #include "cmds/filesystem-usage.h"
 #include "cmds/commands.h"
 #include "kernel-shared/disk-io.h"
-
+#include "common/open-utils.h"
+#include "common/units.h"
 #include "version.h"
 #include "common/help.h"
 #include "common/device-utils.h"
+#include "common/open-utils.h"
 
 /*
  * Add the chunk info to the chunk_info list
@@ -393,6 +396,38 @@ static void get_raid56_space_info(struct btrfs_ioctl_space_args *sargs,
 	}
 }
 
+static u64 get_first_device_zone_size(int fd)
+{
+	int dirfd;
+	DIR *dir;
+	struct dirent *de;
+	char name[NAME_MAX] = {0};
+	u64 ret;
+
+	dirfd = sysfs_open_fsid_dir(fd, "devices");
+	if (dirfd < 0)
+		return 0;
+	dir = fdopendir(dirfd);
+	if (!dir) {
+		ret = 0;
+		goto out;
+	}
+	while (1) {
+		de = readdir(dir);
+		if (strcmp(".", de->d_name) == 0 || strcmp("..", de->d_name) == 0)
+			continue;
+		strcpy(name, de->d_name);
+		name[NAME_MAX - 1] = 0;
+		break;
+	}
+	ret = device_get_zone_size(fd, name);
+	ret *= 512;
+
+out:
+	closedir(dir);
+	return ret;
+}
+
 #define	MIN_UNALOCATED_THRESH	SZ_16M
 static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 		int chunkcount, struct device_info *devinfo, int devcount,
@@ -429,9 +464,11 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 	u64 l_global_reserve_used = 0;
 	u64 free_estimated = 0;
 	u64 free_min = 0;
+	u64 zone_unusable = 0;
 	double max_data_ratio = 1.0;
 	int mixed = 0;
 	struct statfs statfs_buf;
+	struct btrfs_ioctl_feature_flags feature_flags;
 
 	sargs = load_space_info(fd, path);
 	if (!sargs) {
@@ -496,6 +533,16 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 		if ((flags & (BTRFS_BLOCK_GROUP_DATA | BTRFS_BLOCK_GROUP_METADATA))
 		    == (BTRFS_BLOCK_GROUP_DATA | BTRFS_BLOCK_GROUP_METADATA)) {
 			mixed = 1;
+		} else {
+			/*
+			 * As mixed mode is not supported in zoned mode, this
+			 * will account for all profile types
+			 */
+			u64 tmp;
+
+			tmp = device_get_zone_unusable(fd, flags);
+			if (tmp != DEVICE_ZONE_UNUSABLE_UNKNOWN)
+				zone_unusable += tmp;
 		}
 		if (flags & BTRFS_BLOCK_GROUP_DATA) {
 			r_data_used += sargs->spaces[i].used_bytes * ratio;
@@ -575,6 +622,16 @@ static int print_filesystem_usage_overall(int fd, struct chunk_info *chunkinfo,
 		pretty_size_mode(r_total_unused, unit_mode | UNITS_NEGATIVE));
 	printf("    Device missing:\t\t%*s\n", width,
 		pretty_size_mode(r_total_missing, unit_mode));
+	ret = ioctl(fd, BTRFS_IOC_GET_FEATURES, &feature_flags);
+	if (ret == 0 && (feature_flags.incompat_flags & BTRFS_FEATURE_INCOMPAT_ZONED)) {
+		u64 zone_size;
+
+		printf("    Device zone unusable:\t%*s\n", width,
+			pretty_size_mode(zone_unusable, unit_mode));
+		zone_size = get_first_device_zone_size(fd);
+		printf("    Device zone size:\t\t%*s\n", width,
+			pretty_size_mode(zone_size, unit_mode));
+	}
 	printf("    Used:\t\t\t%*s\n", width,
 		pretty_size_mode(r_total_used, unit_mode));
 	printf("    Free (estimated):\t\t%*s\t(",
@@ -712,7 +769,7 @@ static int load_device_info(int fd, struct device_info **device_info_ptr,
 		} else {
 			strcpy(info[ndevs].path, (char *)dev_info.path);
 			info[ndevs].device_size =
-				get_partition_size((char *)dev_info.path);
+				device_get_partition_size((const char *)dev_info.path);
 		}
 		info[ndevs].size = dev_info.total_bytes;
 		++ndevs;
@@ -893,7 +950,7 @@ static void _cmd_filesystem_usage_tabular(unsigned unit_mode,
 			col++;
 		}
 
-		unused = get_partition_size(device_info_ptr[i].path)
+		unused = device_get_partition_size(device_info_ptr[i].path)
 				- total_allocated;
 
 		table_printf(matrix, unallocated_col, vhdr_skip + i, ">%s",
@@ -1173,20 +1230,41 @@ void print_device_chunks(struct device_info *devinfo,
 		const char *r_mode;
 		u64 flags;
 		u64 size;
+		u64 num_stripes;
+		u64 profile;
 
 		if (chunks_info_ptr[i].devid != devinfo->devid)
 			continue;
 
 		flags = chunks_info_ptr[i].type;
+		profile = (flags & BTRFS_BLOCK_GROUP_PROFILE_MASK);
 
 		description = btrfs_group_type_str(flags);
 		r_mode = btrfs_group_profile_str(flags);
 		size = calc_chunk_size(chunks_info_ptr+i);
-		printf("   %s,%s:%*s%10s\n",
-			description,
-			r_mode,
-			(int)(20 - strlen(description) - strlen(r_mode)), "",
-			pretty_size_mode(size, unit_mode));
+		num_stripes = chunks_info_ptr[i].num_stripes;
+
+		switch (profile) {
+		case BTRFS_BLOCK_GROUP_RAID0:
+		case BTRFS_BLOCK_GROUP_RAID5:
+		case BTRFS_BLOCK_GROUP_RAID6:
+		case BTRFS_BLOCK_GROUP_RAID10:
+			printf("   %s,%s/%llu:%*s%10s\n",
+				   description,
+				   r_mode,
+				   num_stripes,
+				   (int)(20 - strlen(description) - strlen(r_mode)
+						 - count_digits(num_stripes) - 1), "",
+				   pretty_size_mode(size, unit_mode));
+			break;
+		default:
+			printf("   %s,%s:%*s%10s\n",
+				   description,
+				   r_mode,
+				   (int)(20 - strlen(description) - strlen(r_mode)), "",
+				   pretty_size_mode(size, unit_mode));
+			break;
+		}
 
 		allocated += size;
 
