@@ -9,15 +9,12 @@ check_prereq btrfs-image
 
 # redefine the one provided by common
 check_image() {
-	local image
-
-	image=$1
 	truncate -s0 target
-	run_mayfail $TOP/btrfs-image "$image" target
+	run_mayfail "$TOP/btrfs-image" "$1" target
 	truncate -s0 target
 }
 
-check_all_images $TEST_TOP/fuzz-tests/images
+check_all_images "$TEST_TOP/fuzz-tests/images"
 
 rm -- target
 
